module Util.ShapeManager exposing (..)
-- Json
import Json.Decode as JD
import Json.Encode
-- Own
import Util.Util exposing (..)
import Url

-- main types
type Shape
  = Empty
  | Circle String String String
  | Rectangle String String String String
  | Ellipse   String String String String
  | Polygon (List ( String, String ) )

type alias ShapeStyle =
  { hoveredFillColor   : String
  , hoveredFillOpacity : String
  , hoveredStrokeColor   : String
  , hoveredStrokeWidth   : String
  , hoveredStrokeOpacity : String
  , unHoveredFillColor   : String
  , unHoveredFillOpacity : String
  , unHoveredStrokeColor   : String
  , unHoveredStrokeWidth   : String
  , unHoveredStrokeOpacity : String
  , title : String
  , body  : String
  }

type ShapesMsg
  = DeleteShape Int
  | AddShape Shape
  | SaveEditedShape Int ShapeStyle
  | SaveShapeStyle  Int ShapeStyle String String

-- Other Types with Shapes
type alias ShapePicture =
    { picture : Maybe Picture
    , shapes  : List (Shape, ShapeStyle)
    }
-- Update
updateShapes : ShapesMsg -> List (Shape, ShapeStyle) -> List (Shape, ShapeStyle)
updateShapes shapesMsg shapeList =
    case shapesMsg of
      DeleteShape id -> List.append (List.take (id-1) shapeList)
                                    (List.drop (id)   shapeList)
      AddShape shape -> List.append shapeList [(shape,defaultShapeStyle)]
      SaveEditedShape id shapeStyle -> List.append (List.take (id-1) shapeList)
                                                   ((shapeList
                                                      |> List.take (id  )
                                                      |> List.drop (id-1)
                                                      |> List.head
                                                      |> Maybe.withDefault (Empty,shapeStyle)
                                                      |> Tuple.first
                                                    ,shapeStyle
                                                    )::(List.drop (id)   shapeList)
                                                   )
      SaveShapeStyle id shapeStyle inputType value -> updateShapes (SaveEditedShape id (case inputType of
                                                                                          "T" -> { shapeStyle | title = value }
                                                                                          "B" -> { shapeStyle | body  = value }
                                                                                          _   -> shapeStyle -- ERROR
                                                                                       )
                                                                   ) shapeList
-- Shape Construktors
shapeDefaultConstruktor : String -> Shape
shapeDefaultConstruktor shape =
    case shape of
      "Circle"   -> Circle    "" "" ""
      "Rectangle"-> Rectangle "" "" "" ""
      "Ellipse"  -> Ellipse "" "" "" ""
      "Polygon"  -> Polygon   []
      _          -> Empty
shapesDecoder : String -> List Shape
shapesDecoder s =
    if (String.contains "shapes" s)
      then []
      else
        List.concat
          [ if (String.contains "circle" s)
              then [ Circle (shapesDecoderList "circle" "x" s)
                            (shapesDecoderList "circle" "y" s)
                            (shapesDecoderList "circle" "r" s)
                   ]
              else []
          , if (String.contains "rectangle" s)
              then [ Rectangle (shapesDecoderList "rectangle" "x"      s)
                               (shapesDecoderList "rectangle" "y"      s)
                               (shapesDecoderList "rectangle" "width"  s)
                               (shapesDecoderList "rectangle" "height" s)
                   ]
              else []
            , if (String.contains "Ellipse" s)
            then [ Ellipse (shapesDecoderList "Ellipse" "cx"      s)
                             (shapesDecoderList "Ellipse" "cy"      s)
                             (shapesDecoderList "Ellipse" "rx"  s)
                             (shapesDecoderList "Ellipse" "ry" s)
                 ]
            else []
          , if (String.contains "polygon" s)
              then [ Polygon (List.map2 Tuple.pair
                             (shapesDecoderListPoly "x" s)
                             (shapesDecoderListPoly "y" s)
                             )
                   ]
              else []
          ]
-- ShapeStyle Constructors
defaultShapeStyle : ShapeStyle
defaultShapeStyle = ShapeStyle "white" "0.5" "red" "3" "1.0"
                               "white" "0.0" "red" "3" "1.0"
                               "default" "default"
-- Operations on Shapes
eq_String_Shape : String -> Shape -> Bool
eq_String_Shape string shape = string==(shapeNameToString shape)
deleteCoordsFromShape : Shape -> Shape
deleteCoordsFromShape shape =
    case shape of
      Empty             -> Empty
      Circle    _ _ _   -> Circle    "" "" ""
      Rectangle _ _ _ _ -> Rectangle "" "" "" ""
      Ellipse   _ _ _ _ -> Ellipse "" "" "" ""
      Polygon   _       -> Polygon   []
isCorrectFilled : Shape -> Bool
isCorrectFilled shape =
    case shape of
      Empty             -> False
      Circle    x y r   -> (isNumeric x)&&(isNumeric y)&&(isNumeric r)
      Rectangle a b c d -> (isNumeric a)&&(isNumeric b)&&(isNumeric c)&&(isNumeric d)
      Ellipse   a b c d -> (isNumeric a)&&(isNumeric b)&&(isNumeric c)&&(isNumeric d)
      Polygon   pList   -> (List.length pList) >= 3
-- Get Methods : Parameter of Shapes
getPointListFromPolygon : Shape -> List (String,String)
getPointListFromPolygon shape =
    case shape of
      Empty               -> []
      Circle    _ _ _     -> []
      Rectangle _ _ _ _   -> []
      Ellipse   _ _ _ _   -> []
      Polygon   pointList -> pointList
-- Transform Shape to other Types
shapeListToShape_ShapeStyleList : List Shape -> List (Shape, ShapeStyle)
shapeListToShape_ShapeStyleList shapes = List.map2 Tuple.pair shapes (List.repeat (List.length shapes) defaultShapeStyle)
shapeNameToString : Shape -> String
shapeNameToString shape =
    case shape of
      Empty             -> ""
      Circle    _ _ _   -> "Circle"
      Rectangle _ _ _ _ -> "Rectangle"
      Ellipse   _ _ _ _ -> "Ellipse"
      Polygon   _       -> "Polygon"
shapeNameStringToGerman : String -> String
shapeNameStringToGerman shape =
    case shape of
      "Circle"   -> "Kreis"
      "Rectangle"-> "Rechteck"
      "Polygon"  -> "Vieleck"
      "Ellipse"  -> "Ellipse"
      _          -> ""
shapeCoordsToString : Shape -> String
shapeCoordsToString shape =
    case shape of
      Empty                                       -> ""
      Circle    x_center y_center radius          -> x_center++","++y_center++","++radius
      Rectangle x_left   y_upper  x_right y_lower -> x_left++","++y_upper++","++x_right++","++y_lower
      Ellipse   x_left   y_upper  x_right y_lower -> x_left++","++y_upper++","++x_right++","++y_lower
      Polygon   pointList                         ->
          case pointList of
            []        -> ""
            (x,y)::xs -> "("++x++","++y++") "++(shapeCoordsToString (Polygon xs))
shapeCoordsToHtmlCoords : Shape -> String
shapeCoordsToHtmlCoords shape =
    case shape of
      Empty             -> ""
      Circle    _ _ _   -> shapeCoordsToString shape
      Rectangle _ _ _ _ -> shapeCoordsToString shape
      Ellipse   _ _ _ _ -> shapeCoordsToString shape
      Polygon   _       -> String.left ((String.length (shapeCoordsToHtmlCoordsHelper shape))-1)
                                       (shapeCoordsToHtmlCoordsHelper shape)
shapeCoordsToHtmlCoordsHelper : Shape -> String
shapeCoordsToHtmlCoordsHelper shape =
    case shape of
      Empty               -> ""
      Circle    _ _ _     -> shapeCoordsToString shape
      Rectangle _ _ _ _   -> shapeCoordsToString shape
      Ellipse   _ _ _ _   -> shapeCoordsToString shape
      Polygon   pointList ->
          case pointList of
            []        -> ""
            (x,y)::xs -> x++","++y++","++(shapeCoordsToHtmlCoordsHelper (Polygon xs))
-- Decoder of shapes
rectListDecoder : JD.Decoder (List Shape)
rectListDecoder = JD.field "rectangle" (JD.list (JD.map4
                                                  Rectangle
                                                    (JD.field "x"      JD.string)
                                                    (JD.field "y"      JD.string)
                                                    (JD.field "width"  JD.string)
                                                    (JD.field "height" JD.string)
                                                )
                                       )
ellipseListDecoder : JD.Decoder (List Shape)
ellipseListDecoder = JD.field "ellipse" (JD.list (JD.map4
                                                  Ellipse
                                                    (JD.field "cx"      JD.string)
                                                    (JD.field "cy"      JD.string)
                                                    (JD.field "rx"  JD.string)
                                                    (JD.field "ry" JD.string)
                                                )
                                       )
shapesDecoderList : String -> String -> String -> String
shapesDecoderList shape var s =
    let res = JD.decodeString
                (JD.field shape (JD.list (JD.field var JD.string)))
                s
    in case res of
         Ok stringList -> case stringList of
                            []       -> ""
                            x::[]    -> x
                            x::xs    -> x++" This List is too long"
         Err e         -> "ERROR"
shapesDecoderListPoly : String -> String -> List String
shapesDecoderListPoly var s =
    let res = JD.decodeString
                (JD.field "polygon" (JD.list (JD.field "points" (JD.list (JD.field var JD.string)))))
                s
    in case res of
         Ok list -> case list of
                      []    -> []
                      x::[] -> x
                      x::xs -> x++["This List is too long"]
         Err e   -> ["ERROR"]
-- new Decoder of Shapes
resultDecode : String -> Result JD.Error ShapePicture
resultDecode jsonString =
    let shapeDecoder : JD.Decoder (List (Shape, ShapeStyle))
        shapeDecoder =
                (JD.list
                  (JD.map2 (\a b -> (a, b))
                    (JD.field ("shape")
                      (JD.oneOf  [

                        (JD.field ( "circle")
                                (JD.map3
                                    (\a b c ->  Circle a b c)
                                    (JD.field "cx" JD.string)
                                    (JD.field "cy" JD.string)
                                    (JD.field "r" JD.string)
                                )
                        ),
                        (JD.field ( "rect")
                                (JD.map4
                                    (\a b c d ->  Rectangle a b c d)
                                    (JD.field "x" JD.string)
                                    (JD.field "y" JD.string)
                                    (JD.field "width" JD.string)
                                    (JD.field "height" JD.string)
                                )
                        ),
                        (JD.field ( "ellipse")
                                (JD.map4
                                    (\a b c d ->  Ellipse a b c d)
                                    (JD.field "cx" JD.string)
                                    (JD.field "cy" JD.string)
                                    (JD.field "rx" JD.string)
                                    (JD.field "ry" JD.string)
                                )
                        ),

                          (JD.map (\l -> Polygon l) (
                          (JD.field ( "polygon")
                              (JD.list
                                  (JD.map2
                                      (\a b -> ( a, b ))
                                      (JD.field "x" JD.string)
                                      (JD.field "y" JD.string)
                                  )
                              )
                          )
                          )
                          )
                          ]
                        )
                      )

                    ((JD.field ("shapeprop") (
                    JD.map4 (\( a, b, c ) ( d, e, f )
                                        ( g, h, i ) ( j, k, l )
                                    -> ShapeStyle a b c d e f g h i j k l)
                        (JD.map3
                        (\a b c -> ( a, b, c ))
                        (JD.field "hFC" JD.string)
                        (JD.field "hFO" JD.string)
                        (JD.field "hSC" JD.string)
                        )
                        (JD.map3 (\a b c -> ( a, b, c ))
                        (JD.field "hSW" JD.string)
                        (JD.field "hSO" JD.string)
                        (JD.field "uFC" JD.string)
                        )
                        (JD.map3 (\a b c -> ( a, b, c ))
                        (JD.field "uFO" JD.string)
                        (JD.field "uSC" JD.string)
                        (JD.field "uSW" JD.string)
                        )
                        (JD.map3 (\a b c -> ( a, b, c ))
                        (JD.field "uSO" JD.string)
                        (JD.field "title" JD.string)
                        (JD.field "body" JD.string)
                      )
                   ))
                   )
                   )
                  )
        shapePictureDecoder : JD.Decoder ShapePicture
        shapePictureDecoder = JD.map2
          (\a b ->
            let
              maybePicture = Maybe.map (\u -> Picture u 0 0) (Url.fromString a)
            in
            ShapePicture maybePicture b
          )
          ( JD.field "pictureUrl" JD.string)
          ( JD.field "shapes"     shapeDecoder      )
    in JD.decodeString shapePictureDecoder jsonString
-- new Encoder of Shapes
writeAlltoJson : List ( Shape, ShapeStyle ) -> String -> Json.Encode.Value
writeAlltoJson shapeWithProperties picUrl =
  let
      jsonObj =
        Json.Encode.object
        ( [( "pictureUrl", Json.Encode.string picUrl ), ("shapes", writeShapeToJson shapeWithProperties )])
  in jsonObj

writeShapeToJson : List ( Shape, ShapeStyle ) -> Json.Encode.Value
writeShapeToJson data =
    Json.Encode.list shapeCoordsToJson data

shapeCoordsToJson : ( Shape, ShapeStyle ) -> Json.Encode.Value
shapeCoordsToJson shapeWithProperties =
    case Tuple.first shapeWithProperties of
      Empty                                       -> Json.Encode.string ""
      Circle    x_center y_center radius          -> Json.Encode.object ( [ ("shape",  circToJson x_center y_center radius ), ("shapeprop", shapePropertiestoJson (Tuple.second shapeWithProperties))])
      Rectangle x_left   y_upper  x_right y_lower -> Json.Encode.object ( [ ("shape",  rectToJson x_left y_upper x_right y_lower ), ("shapeprop", shapePropertiestoJson (Tuple.second shapeWithProperties))])
      Ellipse   x_left   y_upper  x_right y_lower -> Json.Encode.object ( [ ("shape",  ellipseToJson x_left y_upper x_right y_lower ), ("shapeprop", shapePropertiestoJson (Tuple.second shapeWithProperties))])
      Polygon   pointList                         -> Json.Encode.object ( [ ("shape",  polyToJson pointList ), ("shapeprop", shapePropertiestoJson (Tuple.second shapeWithProperties))])

circToJson : String -> String -> String -> Json.Encode.Value
circToJson a b c =
  let
      jsonObj =
        Json.Encode.object
          [
          (
          "circle"
          ,Json.Encode.object ([ ( "cx", Json.Encode.string a ),
                                 ( "cy", Json.Encode.string b ),
                                 ( "r", Json.Encode.string c ) ])
          )
          ]
  in jsonObj

shapePropertiestoJson : ShapeStyle -> Json.Encode.Value
shapePropertiestoJson shapestyle =
      let
          jsonObj =

              Json.Encode.object ([ ( "hFC", Json.Encode.string shapestyle.hoveredFillColor ),
                                     ( "hFO", Json.Encode.string shapestyle.hoveredFillOpacity ),
                                     ( "hSC", Json.Encode.string shapestyle.hoveredStrokeColor ),
                                     ( "hSW", Json.Encode.string shapestyle.hoveredStrokeWidth ),
                                     ( "hSO", Json.Encode.string shapestyle.hoveredStrokeOpacity ),
                                     ( "uFC", Json.Encode.string shapestyle.unHoveredFillColor ),
                                     ( "uFO", Json.Encode.string shapestyle.unHoveredFillOpacity ),
                                     ( "uSC", Json.Encode.string shapestyle.unHoveredStrokeColor ),
                                     ( "uSW", Json.Encode.string shapestyle.unHoveredStrokeWidth ),
                                     ( "uSO", Json.Encode.string shapestyle.unHoveredStrokeOpacity ),
                                     ( "title", Json.Encode.string shapestyle.title ),
                                     ( "body", Json.Encode.string shapestyle.body )
                                  ])
      in jsonObj

polyToJson : List ( String, String ) -> Json.Encode.Value
polyToJson data =
    let
        jsonObj =
            Json.Encode.object
                [ (
                  "polygon"
                  , Json.Encode.list (\( i, j ) ->
                    Json.Encode.object [
                    ( "x", Json.Encode.string i ), ( "y", Json.Encode.string j ) ]
                    ) data
                  )
                ]
    in jsonObj
polyList2Json : List (List ( String, String )) -> Json.Encode.Value
polyList2Json data =
    Json.Encode.list polyToJson data

ellipseToJson : String -> String -> String -> String -> Json.Encode.Value
ellipseToJson a b c d =
  let
      jsonObj =
        Json.Encode.object
            [
            (
            "ellipse"
            ,Json.Encode.object ([ ( "cx", Json.Encode.string a ),
                                   ( "cy", Json.Encode.string b ),
                                   ( "rx", Json.Encode.string c ),
                                   ( "ry", Json.Encode.string d ) ])
            )
            ]
  in jsonObj

rectToJson : String -> String -> String -> String -> Json.Encode.Value
rectToJson a b c d =
  let
      jsonObj =
        Json.Encode.object
          [
          (
          "rect"
          ,Json.Encode.object ([ ( "x", Json.Encode.string a ),
                                 ( "y", Json.Encode.string b ),
                                 ( "width", Json.Encode.string c ),
                                 ( "height", Json.Encode.string d ) ])
          )
          ]
  in jsonObj
